/**HEADER********************************************************************
*
* Copyright (c) 2008 Freescale Semiconductor;
* All Rights Reserved
*
***************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: httpd.c$
* $Version : 3.8.33.0$
* $Date    : Sep-21-2012$
*
* Comments:
*
*   This file contains the HTTPD implementation.
*
*END************************************************************************/

#include "httpd.h"
#include "httpd_supp.h"
#include "httpd_prv.h"
#include "rtcs_base64.h"
#include <string.h>
#include <stdlib.h>

#define MAX(a, b)       (((a) > (b)) ? (a) : (b))

const HTTPD_ROOT_DIR_STRUCT default_root_dir[] = {
    { "", "tfs:" },
    { 0, 0 }
};

static uint_32 httpd_init_socket (HTTPD_STRUCT *server, uint_16 family);
static uint_32 httpd_set_params  (HTTPD_STRUCT *server, HTTPD_PARAMS_STRUCT* params);
/*
** Function for starting the HTTP server 
**
** IN:
**      HTTPD_PARAMS_STRUCT*   params - server parameters (port, ip, index page etc.)
**
** OUT:
**      none
**
** Return Value: 
**      uint_32      server handle if successful, NULL otherwise
*/
uint_32 httpd_init(HTTPD_PARAMS_STRUCT *params)
{
    HTTPD_STRUCT *server = NULL;
    uint_32 error;
    uint_32 error4 = HTTPD_OK;
    uint_32 error6 = HTTPD_OK;

    
    if ((server = _mem_alloc_zero(sizeof(HTTPD_STRUCT))) == NULL)
    {
        HTTPD_DEBUG(1, "server: main structure allocation error\n");
        return((uint_32) NULL);
    }
    _mem_set_type(server, MEM_TYPE_HTTPD_STRUCT);

    error = _lwsem_create(&server->tid_sem, 1);
    if (error != MQX_OK)
    {
        httpd_release((uint_32) server);
        return((uint_32) NULL);
    }

    error = httpd_set_params(server, params);
    if (error != HTTPD_OK)
    {
        httpd_release((uint_32) server);
        return((uint_32) NULL);
    }
    
    /* Allocate space for session pointers */
    server->session = _mem_alloc_zero(sizeof(HTTPD_SESSION_STRUCT*) * server->params.max_ses);
    if (server->session == NULL)
    {
        return(MQX_OUT_OF_MEMORY);
    }

    /* Allocate space for session task IDs */
    server->ses_tid = _mem_alloc_zero(sizeof(_rtcs_taskid) * server->params.max_ses);
    if (server->ses_tid == NULL)
    {
        httpd_release((uint_32) server);
        return(MQX_OUT_OF_MEMORY);
    }

    /* Init sockets. */
    if (params->af & AF_INET)
    {
        /* Setup IPv4 server socket */
        error4 = httpd_init_socket(server, AF_INET);
    }
    if (params->af & AF_INET6)
    {
        /* Setup IPv6 server socket */
        error6 = httpd_init_socket(server, AF_INET6);
    }

    if ((error4 != HTTPD_OK) || (error6 != HTTPD_OK))
    {
        HTTPD_DEBUG(1,"IPv4 Error 0x%X\n", error4);
        HTTPD_DEBUG(1,"IPv6 Error 0x%X\n", error6);
        httpd_release((uint_32) server);
        return((uint_32) NULL);
    }
    server->run = 1;
    HTTPD_DEBUG(1, "------------------------------------------\n");
    HTTPD_DEBUG(1, " HTTP 1.0 server has been started.\n");
    HTTPD_DEBUG(1, "------------------------------------------\n\n");

    if (RTCS_task_create("httpd server", server->params.server_prio, server->params.server_stack, httpd_server_task, server) != RTCS_OK)
    {
        httpd_release((uint_32) server);
        return((uint_32) NULL);
    }

    return((uint_32) server);
}

/*
** Function for releasing/stopping HTTP server 
**
** IN:
**      uint_32       server_h - server handle
**
** OUT:
**      none
**
** Return Value: 
**      uint_32      error code. HTTPD_OK if everything went right, positive number otherwise
*/
uint_32 httpd_release(uint_32 server_h)
{
    uint_32 i;
    uint_32 n = 0;
    HTTPD_STRUCT* server = (pointer) server_h;
    boolean wait = FALSE;

    /* Shutdown server listen sockets, stop server task */
    if (server != NULL)
    {
        if (server->sock_v4)
        {
            shutdown(server->sock_v4, FLAG_ABORT_CONNECTION);
        }
            
        if (server->sock_v6)
        {
            shutdown(server->sock_v6, FLAG_ABORT_CONNECTION);
        }
    }
    else
    {
        return(HTTPD_ERR);
    }

    /* Invalidate sessions (this is signal for session tasks to end them) */
    while(n < server->params.max_ses)
    {
        if (server->session[n])
        {
            server->session[n]->valid = HTTPD_SESSION_INVALID;
        }
        n++;
    }
    /* Wait until all session tasks end */
    do
    {
        wait = FALSE;
        for (n = 0; n < server->params.max_ses; n++)
        {
            if (server->ses_tid[n])
            {
                wait = TRUE;
            }
        }
        _sched_yield();
    }while(wait);

    /* Shutdown server task */
    server->run = 0;
    
    while(server->server_tid)
    {
        _sched_yield();
    }

    _lwsem_destroy(&server->tid_sem);
    
    /* Free memory */
    _mem_free(server->ses_tid);
    server->ses_tid = NULL;
    _mem_free(server->session);
    server->session = NULL;
    _mem_free(server);
    return(HTTPD_OK);
}

/*
** Internal function for server parameters initialization
**
** IN:
**      HTTPD_STRUCT* server - server structure pointer
**
**      HTTPD_PARAMS_STRUCT* params - pointer to user parameters if there are any
** OUT:
**      none
**
** Return Value: 
**      uint_32      error code. HTTPD_OK if everything went right, positive number otherwise
*/
static uint_32 httpd_set_params(HTTPD_STRUCT *server, HTTPD_PARAMS_STRUCT *params)
{
    /* Stack size */
    uint_32 stack;

    server->params.port = HTTPDCFG_DEF_PORT;
    server->params.ipv4_address.s_addr = HTTPDCFG_DEF_ADDR;
#if RTCSCFG_ENABLE_IP6  
    server->params.ipv6_address = in6addr_any;
    server->params.ipv6_scope_id = 0;
#endif
    server->params.af = AF_INET;
    server->params.max_uri = HTTPDCFG_DEF_URL_LEN;
    server->params.max_auth = HTTPDCFG_DEF_AUTH_LEN;
    server->params.max_ses = HTTPDCFG_DEF_SES_CNT;
    stack = HTTPD_SERVER_STACK_SIZE;
    server->params.server_stack = stack;
    server->params.root_dir = (HTTPD_ROOT_DIR_STRUCT*) &default_root_dir;
    server->params.index_page = HTTPDCFG_DEF_INDEX_PAGE;
    server->params.cgi_lnk_tbl = NULL;
    server->params.fn_lnk_tbl = NULL;
    server->params.auth_fn = NULL;
    server->params.server_prio = HTTPDCFG_DEF_SERVER_PRIO;
    server->params.ses_prio = HTTPDCFG_DEF_SERVER_PRIO;
    stack = MAX(MAX(HTTPD_MAX_LEN, 512), HTTPDCFG_MAX_SCRIPT_LN) + 10 * sizeof(void*);
    stack = MAX(3 * stack, HTTPD_MIN_SES_STACK_SIZE);
    server->params.ses_stack = stack;

    /* If there is parameters structure copy nonzero values to server */
    // ADD ifdef for IPv4
    if (params != NULL)
    {
        if (params->port)
            server->params.port = params->port;
        if (params->ipv4_address.s_addr != 0)
            server->params.ipv4_address = params->ipv4_address;
    #if RTCSCFG_ENABLE_IP6
        if (params->ipv6_address.s6_addr != NULL)
            server->params.ipv6_address = params->ipv6_address;
        if (params->ipv6_scope_id)
            server->params.ipv6_scope_id = params->ipv6_scope_id;
    #endif
        if (params->af)
            server->params.af = params->af;
        if (params->max_uri)
            server->params.max_uri = params->max_uri;
        if (params->max_auth)
            server->params.max_auth = params->max_auth;
        if (params->max_ses)
            server->params.max_ses = params->max_ses;
        if (params->root_dir)
            server->params.root_dir = params->root_dir;
        if (params->index_page)
            server->params.index_page = params->index_page;
        if (params->cgi_lnk_tbl)
            server->params.cgi_lnk_tbl = params->cgi_lnk_tbl;
        if (params->fn_lnk_tbl)
            server->params.fn_lnk_tbl = params->fn_lnk_tbl;
        if (params->auth_fn)
            server->params.auth_fn = params->auth_fn;
        if (params->server_prio)
            server->params.server_prio = params->server_prio;
        if (params->server_stack)
            server->params.server_stack = params->server_stack;
        if (params->ses_prio)
            server->params.ses_prio = params->ses_prio;
        if (params->ses_stack)
            server->params.ses_stack = params->ses_stack;
    }

    return(HTTPD_OK);
}

/*
** Function for socket initialization (both IPv4 and IPv6)
**
** IN:
**      HTTPD_STRUCT* server - server structure pointer
**
**      uint_16      family - IP protocol family
** OUT:
**      none
**
** Return Value: 
**      uint_32      error code. HTTPD_OK if everything went right, positive number otherwise
*/
static uint_32 httpd_init_socket(HTTPD_STRUCT *server, uint_16 family)
{
    uint_32 option;
    uint_32 error;
    sockaddr sin_sock;
    uint_32 sock = 0;

    _mem_zero(&sin_sock, sizeof(sockaddr));

    if (family == AF_INET) /* IPv4 */
    {
        #if RTCSCFG_ENABLE_IP4
        if ((server->sock_v4 = socket(AF_INET, SOCK_STREAM, 0)) == RTCS_ERROR)
        {
            return(HTTPD_CREATE_FAIL);
        }
        ((sockaddr_in *)&sin_sock)->sin_port   = server->params.port;
        ((sockaddr_in *)&sin_sock)->sin_addr   = server->params.ipv4_address;
        ((sockaddr_in *)&sin_sock)->sin_family = AF_INET;
        sock = server->sock_v4;
        #else
        return(HTTPD_BAD_FAMILY);
        #endif
    }
    else if (family == AF_INET6) /* IPv6 */
    {
        #if RTCSCFG_ENABLE_IP6
        if ((server->sock_v6 = socket(AF_INET6, SOCK_STREAM, 0)) == RTCS_ERROR)
        {
            return(HTTPD_CREATE_FAIL);
        }
        ((sockaddr_in6 *)&sin_sock)->sin6_port      = server->params.port;
        ((sockaddr_in6 *)&sin_sock)->sin6_family    = AF_INET6;
        ((sockaddr_in6 *)&sin_sock)->sin6_scope_id  = server->params.ipv6_scope_id;
        ((sockaddr_in6 *)&sin_sock)->sin6_addr      = server->params.ipv6_address;
        sock = server->sock_v6;
        #else
        return(HTTPD_BAD_FAMILY);
        #endif
    }
    else
    {
        return(HTTPD_BAD_FAMILY);
    }
    /* Set socket options */
    option = 500;
    error = setsockopt(sock, SOL_TCP, OPT_SEND_TIMEOUT, &option, sizeof(option));
    option = TRUE;
    error = setsockopt(sock, SOL_TCP, OPT_SEND_NOWAIT, &option, sizeof(option));
    option = 5000;
    error = setsockopt(sock, SOL_TCP, OPT_CONNECT_TIMEOUT, &option, sizeof(option));
    option = HTTPCFG_TIMEWAIT_TIMEOUT;
    error = setsockopt(sock, SOL_TCP, OPT_TIMEWAIT_TIMEOUT, &option, sizeof(option));
    option = 50;     
    error = setsockopt(sock, SOL_TCP, OPT_RECEIVE_TIMEOUT, &option, sizeof(option));
    option = FALSE; 
    error = setsockopt(sock, SOL_TCP, OPT_RECEIVE_NOWAIT, &option, sizeof(option));
    option = TRUE;  
    error = setsockopt(sock, SOL_TCP, OPT_RECEIVE_PUSH, &option, sizeof(option));
    option = TRUE;
    error = setsockopt(sock, SOL_TCP, OPT_NO_NAGLE_ALGORITHM, &option, sizeof(option));
    option = HTTPCFG_TX_BUFFER_SIZE;
    error = setsockopt(sock, SOL_TCP, OPT_TBSIZE, &option, sizeof(option));
    option = HTTPCFG_RX_BUFFER_SIZE;
    error = setsockopt(sock, SOL_TCP, OPT_RBSIZE, &option, sizeof(option));

    if (error != RTCS_OK)
    {
        HTTPD_DEBUG(1, "socket: bind error\n");
        return(HTTPD_SOCKOPT_FAIL);
    }

    /* Bind socket */
    error = bind(sock, &sin_sock, sizeof(sin_sock));
    if(error != RTCS_OK)
    {
        HTTPD_DEBUG(1, "socket: bind error\n");
        return(HTTPD_BIND_FAIL);
    }

    /* Listen */
    error = listen(sock, 0);
    if (error != RTCS_OK)
    {
        HTTPD_DEBUG(1, "socket: listen error\n");
        return(HTTPD_LISTEN_FAIL);
    }
    return(HTTPD_OK);
}

/*
** Function used to free session structure
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer
**
** OUT:
**      none
**
** Return Value: 
**      none
*/
void httpd_ses_free(HTTPD_SESSION_STRUCT *session)
{
    HTTPD_ASSERT(session);

    if (session)
    {
        if (session->request.path)
        {
            _mem_free(session->request.path);
            session->request.path = NULL;
        }
        if(session->request.auth)
        {
            _mem_free(session->request.auth);
            session->request.auth = NULL; 
        }
        if(session->buffer.data)
        {
            _mem_free(session->buffer.data);
            session->buffer.data = NULL;
        }
        _mem_free(session);
        session = NULL;
    }
}

/*
** Function for session allocation
**
** IN:
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**
** OUT:
**      none
**
** Return Value: 
**      HTTPD_SESSION_STRUCT* - pointer to allocated session. Non-zero if allocation was OK, NULL otherwise
*/
HTTPD_SESSION_STRUCT* httpd_ses_alloc(HTTPD_STRUCT *server)
{
    HTTPD_SESSION_STRUCT *session = NULL;

    HTTPD_ASSERT(server);

    if (server)
    {
        session = _mem_alloc_zero(sizeof(HTTPD_SESSION_STRUCT));
        if (session)
        {
            _mem_set_type(session, MEM_TYPE_HTTPD_SESSION_STRUCT);

            session->request.path = _mem_alloc_zero(server->params.max_uri + 1);
            if (NULL == session->request.path)
            {
                goto ERROR;
            }
               
            _mem_set_type(session->request.path, MEM_TYPE_HTTPD_URI);

            session->buffer.data = _mem_alloc_zero(sizeof(char)*HTTPDCFG_SES_BUFFER_SIZE);
            if (NULL == session->buffer.data)
            {
                goto ERROR;
            }
        }
    }

    return session;

ERROR:
    if (session->request.path)
    {
        _mem_free(session->request.path);
        session->request.path = NULL;
    }
    if (session->buffer.data)
    {
        _mem_free(session->buffer.data);
        session->buffer.data = NULL;
    }
    _mem_free(session);
    return NULL;
}

/*
** Function used to init session structure
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters)
**      const int sock - socket handle used for communication with client
**
** OUT:
**      none
**
** Return Value: 
**      none
*/
void httpd_ses_init(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session, const int sock)
{
    HTTPD_ASSERT(server && session && sock != RTCS_ERROR);

    HTTPD_DEBUG(5, "httpd_ses_init: server(%p) session(%p) sock(0x%x)\n", server, session, sock);

    if (server && session)
    {
        // init session
        session->state = HTTPD_SES_WAIT_REQ;
        session->sock = sock;
        session->keep_alive = 0;
        session->valid = HTTPD_SESSION_VALID;
        session->request.process_header = 1;
        session->timeout = HTTPDCFG_SES_TO;
        session->keep_alive_enabled = HTTPDCFG_KEEPALIVE_ENABLED;
        session->start_time = RTCS_time_get();
    }
}

/*
** Function used to close session
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer
**
** OUT:
**      none
**
** Return Value: 
**      none
*/
void httpd_ses_close(HTTPD_SESSION_STRUCT *session)
{
    HTTPD_ASSERT(session);
    HTTPD_DEBUG(1, "session closed (%p)\n", session);

    if (session != NULL)
    {
        if (session->response.file)
        {
            fclose(session->response.file);
            session->response.file = NULL;
        }
        session->valid = HTTPD_SESSION_INVALID;
        shutdown(session->sock, FLAG_CLOSE_TX);
    }
}
/*
** Process one line of http request header
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer.
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**      char* buffer - pointer to begining of line with request.
**
** OUT:
**      none
**
** Return Value:
**      none
*/
static void httpd_process_req_hdr_line(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT* session, char* buffer)
{
    char* param_ptr = NULL;

    if (strncmp(buffer, "Connection: ", 12) == 0)
    {
        param_ptr = buffer+12;

        if (session->keep_alive_enabled && (strncmp(param_ptr, "keep-alive", 10) == 0) || (strncmp(param_ptr, "Keep-Alive", 10) == 0))
        {
            /* You should keep in mind that HTTP 1.0 does not oficialy support keep alive */
            /* This is only work around */
            session->keep_alive = 1;
            HTTPD_DEBUG(4, "Connection: keep-alive = 1\n");
        }
    }
    else if (strncmp(buffer, "Content-Length: ", 16) == 0)
    {
        param_ptr = buffer+16;
        session->request.content_len = strtoul(param_ptr, NULL, 10);
        HTTPD_DEBUG(4, "Content-Length: %u\n", session->request.content_len);
    }
    else if (strncmp(buffer, "Authorization: ", 15) == 0)
    {
        param_ptr = buffer+15;
        if (strncmp(param_ptr, "Basic ", 6) == 0)
        {
            uint_32 decoded_length = 0;
            param_ptr += 6;
            /* evaluate number of bytes required for worst case (no padding) */
            decoded_length = (strlen(param_ptr)/4) * 3 + 1;
            session->request.auth = _mem_alloc_zero(sizeof(char)*decoded_length);
            if (session->request.auth != NULL)
            {
                _mem_set_type(session->request.auth, MEM_TYPE_HTTPD_AUTH);
                base64_decode(session->request.auth, param_ptr, decoded_length);
            }
            else
            {
                return;
            }

            param_ptr = strchr(session->request.auth, ':');
            if (param_ptr)
            {
                *param_ptr = '\0';
                session->request.authPasswd = param_ptr + 1;
            }

            HTTPD_DEBUG(4, "Authorization - Basic: %s:%s\n", session->request.auth, session->request.authPasswd);
        }
        else
        {
            HTTPD_DEBUG(4, "Authorization - unknown method\n");
        }
    }
}
/*
** Read http method 
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer.
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**      char* buffer - pointer to begining of line with request.
**
** OUT:
**      none
**
** Return Value:
**      none
*/
static void httpd_process_req_method(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session, char* buffer)
{
    char* uri_begin  = NULL;
    char* uri_end = NULL;
    char method[5] = {'\0'};
    uint_32 written;

    snprintf(method, sizeof(method), "%s", buffer);
    if (strncmp(method, "GET", 3) == 0)
    {
        session->request.method = HTTPD_REQ_GET;
    }
    else if (strncmp(method, "POST", 4) == 0)
    {
        session->request.method = HTTPD_REQ_POST;
    }
    else // unknown method - not implemented response
    {  
        session->request.method = HTTPD_REQ_UNKNOWN;
        return;
    }

    // parse remaining part of line
    uri_begin = strchr(buffer, ' ') + 1;

    if (uri_begin != NULL)
    {
        uri_end = strchr(uri_begin, ' ');
        if (uri_end != NULL)
        {
            *uri_end = '\0';
        }
    }

    written = snprintf(session->request.path, server->params.max_uri, "%s", uri_begin);
    /* Check if whole URI is saved in buffer */
    if (written > server->params.max_uri-1)
    {
        session->request.path[0] = '\0';
        /* URI is too long so we set proper status code for response */
        session->response.statuscode = 414;
    }
}

/*
** Function for request parsing
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer.
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**
** OUT:
**      none
**
** Return Value: 
**      int - zero if request is valid, negative value if invalid.
*/
static int httpd_readreq(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session)
{
    char* buf = session->buffer.data;
    char *cp, *cp2;
    int res;
    uint_32 req_lines = 0;    
    char* end = NULL;
    
    HTTPD_ASSERT(server && session);

    HTTPD_DEBUG(4, "request BEGIN\n");

    /* Begin */
    while (session->request.process_header)
    {
        /* Read data */
        res = recv(session->sock, buf+session->buffer.offset, HTTPDCFG_SES_BUFFER_SIZE-1, 0);
        if ((res == RTCS_ERROR) || (res == 0))
        {
            break;
        }
        /* Process data */
        while (session->request.process_header && ((end = strchr(buf, '\n')) != NULL))
        {
            *end = '\0';
            if (*(end-1) == '\r')
            {
                *(end-1) = '\0';
            }
            req_lines++;
            end++;
            /* Found an empty line => end of header */
            if (!strlen(buf))
            {
                session->request.process_header = FALSE;
            }
            else
            {
                if (req_lines == 1) 
                {
                    httpd_process_req_method(server, session, buf);
                }
                else
                {
                    httpd_process_req_hdr_line(server, session, buf);
                }
            }
            buf = end;
        }
        session->buffer.offset = strlen(buf);
        /* We are not at the end of buffer and there are some data after the end of last line. */
        if (session->buffer.offset != 0)
        {
            /* Copy rest of data to begining of buffer and save offset for next reading. */
            memmove(session->buffer.data, buf, session->buffer.offset);
            buf = session->buffer.data;
        }
        /* Clear the buffer so we dont have some old data there. */
        _mem_zero(session->buffer.data+session->buffer.offset, HTTPDCFG_SES_BUFFER_SIZE-session->buffer.offset);
    }
    /* End */

    // process any URL data
    cp = strchr(session->request.path, '?');
    if (0 != cp)
    {
        *cp = 0;
        cp++;
        session->request.urldata = cp;
    }

    HTTPD_DEBUG(4, "request END res = %d\n", res);

    return (0 > res) ? res : 0;
}

/*
** Function for request processing
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer.
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**
** OUT:
**      none
**
** Return Value: 
**      none
*/
static void httpd_processreq(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session)
{
    char buf[HTTPD_MAX_LEN + 1];
    char *cp, *cp1;
    int cgi = 0;
    int res = 0;
    
    HTTPD_ASSERT(server && session);

    session->state = HTTPD_SES_RESP;

    /* If method is not implemented return without request processing */
    if (session->request.method == HTTPD_REQ_UNKNOWN)
    {
        session->response.statuscode = 501;
    }
    /* We need contetnt length for post requests */
    if ((session->request.method == HTTPD_REQ_POST) && (session->request.content_len == 0))
    {
        session->response.statuscode = 411;
    }
    /* Check auth */
    if (session->request.auth && server->params.auth_fn(session))
    {
        session->response.statuscode = 401;
    }
    /* If response status is not already set than set it to default value */
    if (session->response.statuscode == 0)
    {
        session->response.statuscode = 200;
    }
    else /* We already know how to respond and response is not going to be file */
    {
        return;
    }
    
    /* Check request path */
    cp = strrchr(session->request.path, '/');
    if (0 == cp)
    {
        /* We have invalid request */
        session->response.statuscode = 400;
        HTTPD_DEBUG(1, "Invalid request path '%s'\n", session->request.path);
        return;
    }

    if (session->request.path[0] == '/' && session->request.path[1] == '\0')
    {
        snprintf(session->request.path, server->params.max_uri + 1, "%s", server->params.index_page);
    }

    // check if request is script request

    // if request contain data
    res = -1;

    if ((cp1 = strrchr(cp, '.')) != 0)
    {
        if (0 == strcasecmp(cp1, ".cgi"))   // file extension is cgi, maybe cgi script
        {
            cgi = 1;

            // search cgi script
            *cp1 = 0;
            res = httpd_call_fn(server->params.cgi_lnk_tbl, session, cp + 1);   // cp + 1 = cgi script name begin
            *cp1 = '.';
        }
    }

    if (cgi || session->request.content_len)
    {
        if (res < 0)
        {
            if (res != -2)
            {
                cgi = 0;
            }
            
            // cgi script not found - read data from socket - discard it
            while (session->request.content_len)
            {
                res = httpd_read(session, buf, (int)((res > sizeof(buf))? sizeof(buf) : session->request.content_len));

                session->request.content_len -= res;

                // temporary solution
                if (res == 0)
                {
                    break;
                }  
            }
        }
        else
        {
            // cgi script found - decrease content len
            session->request.content_len -= res;

            if (session->request.content_len == 0)  // whole content procesed
            {
                session->state = HTTPD_SES_END_REQ;
            }    
        }
    }

    // send regular file(document) - if it is not cgi script
    if (!cgi)
    {
         HTTPD_ROOT_DIR_STRUCT *root_dir;
        // get correct root dir (by alias)

        if (server->params.root_dir[1].alias)
        {
            // more root dir pressents, find alias
            root_dir = server->params.root_dir;

            cp = strchr(session->request.path, '/');

            if (cp)
            {
                cp++;
                cp1 = strchr(cp, '/');

                if (cp1)
                {
                    while (root_dir->alias)
                    {
                        if (strncmp(root_dir->alias, cp, cp1 - cp) == 0)
                        {
                            snprintf(buf, HTTPD_MAX_LEN+1, "%s", root_dir->path);
                            strcat(buf, cp1);
                            break;
                        }
                        root_dir++;
                    }
                }
            }

            if (!cp || !cp1 || !root_dir->alias)
            {
                snprintf(buf, HTTPD_MAX_LEN+1, "%s", server->params.root_dir->path);
                strcat(buf, session->request.path);
            }
        }
        else
        {
            snprintf(buf, HTTPD_MAX_LEN+1, "%s", server->params.root_dir->path);
            strcat(buf, session->request.path);
        }

        // correct slash
        cp = buf;
        while (*cp)
        {
            if ('/' == *cp)
            {
                *cp = '\\';
            }
            cp++;
        }

        session->response.file = fopen(buf, "rb");
        session->response.len = 0;

        if (!session->response.file)
        {
            session->response.statuscode = 404;
        }
    }
}

/*
** Function for HTTP sending response
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer.
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**
** OUT:
**      none
**
** Return Value: 
**      none
*/

static void httpd_response(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session)
{
    HTTPD_ASSERT(server && session);
    
    switch (session->response.statuscode)
    {
        case 200:
            httpd_sendfile(server, session);
            break;
        case 400:
            session->keep_alive = 0;
            session->state = HTTPD_SES_END_REQ;
            httpd_sendhdr(session, 0, 0);
            break;
        case 401:
            session->state = HTTPD_SES_END_REQ;
            httpd_send_err_page(session, "401 Unauthorized", "Unauthorized!");
            break;
        case 403:
            session->state = HTTPD_SES_END_REQ;
            httpd_send_err_page(session, "403 Forbidden", "Forbidden!");
            break;
        case 411:
            session->keep_alive = 0;
            session->state = HTTPD_SES_END_REQ;
            httpd_sendhdr(session, 0, 0);
            break;
        case 414:
            session->keep_alive = 0;
            session->state = HTTPD_SES_END_REQ;
            httpd_send_err_page(session, "414 Uri too long", "Requested URI too long!");
            break;
        case 501:
            session->keep_alive = 0;
            session->state = HTTPD_SES_END_REQ;
            httpd_sendhdr(session, 0, 0);
            break;
        default:
            session->state = HTTPD_SES_END_REQ;
            httpd_send_err_page(session, "404 Not Found", "Requested URL not found!");
            break;
    }
}
/*
** HTTP session state machine
**
** IN:
**      HTTPD_SESSION_STRUCT* session - session structure pointer.
**      HTTPD_STRUCT *server - pointer to server structure (needed for session parameters).
**
** OUT:
**      none
**
** Return Value: 
**      none
*/
void httpd_ses_process(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session)
{
    uint_32 time_now;
    int res;
    HTTPD_ASSERT(server && session);

    if (HTTPD_SESSION_VALID == session->valid)
    {
        /* check session timeout */
        time_now = RTCS_time_get();
        if (RTCS_timer_get_interval(session->start_time, time_now) > session->timeout)
        {
            HTTPD_DEBUG(1, "session %p timeout\n", session);
            session->state = HTTPD_SES_CLOSE;
        }

        switch (session->state)
        {
        case HTTPD_SES_WAIT_REQ:
            res = httpd_readreq(server, session);
            
            if ((res < 0) && (!session->keep_alive))
            {
                session->state = HTTPD_SES_END_REQ;
            }
            else if ((res == 0) && (session->request.process_header == 0))
            {
                session->state = HTTPD_SES_PROCESS_REQ;
            }
            break;
        case HTTPD_SES_PROCESS_REQ:
            httpd_processreq(server, session);
            break;

        case HTTPD_SES_RESP:
            httpd_response(server, session);
            break;

        case HTTPD_SES_END_REQ:
            if (!session->keep_alive)
            {
                session->state = HTTPD_SES_CLOSE;
            }
            else
            {
                /* Re-init session */
                session->state = HTTPD_SES_WAIT_REQ;
                if (session->response.file)
                {
                    fclose(session->response.file);
                    /* To be sure that file handle is not used. */
                    session->response.file = NULL;
                }
                session->request.process_header = 1;
                session->start_time = RTCS_time_get();
                session->timeout = HTTPDCFG_KEEPALIVE_TO;
                session->keep_alive = 0;
                break;
            }

        case HTTPD_SES_CLOSE:
            httpd_ses_close(session);
            break;

        default:
            // invalid state
            session->valid = HTTPD_SESSION_INVALID;
            break;
        }
    }
}

