/**HEADER********************************************************************
* 
* Copyright (c) 2008 Freescale Semiconductor;
* All Rights Reserved                       
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: httpd_supp.c$
* $Version : 3.8.28.0$
* $Date    : Aug-31-2012$
*
* Comments:
*
*   HTTPD support functions.
*
*END************************************************************************/

#include "httpd.h"
#include <string.h>
#include <fio.h>
#include <posix.h>

/** string table item
 */
typedef struct {
    int id;
    char *str;
} tbl_str_itm_t;

/* Structure defining MIME types table row */
typedef struct
{
    /* extension length (strlen) */
    uint_32 length;
    /* extension string */
    char* ext;
    /* extension MIME type string */
    int content_type;
}httpd_content_table_row;

/** content type
 */
static const tbl_str_itm_t content_type[] = {
        { CONTENT_TYPE_PLAIN, "text/plain" },
        { CONTENT_TYPE_HTML, "text/html" },
        { CONTENT_TYPE_CSS, "text/css" },
        { CONTENT_TYPE_GIF, "image/gif" },
        { CONTENT_TYPE_JPG, "image/jpeg" },
        { CONTENT_TYPE_PNG, "image/png" },
        { CONTENT_TYPE_JS, "application/javascript"},
        { CONTENT_TYPE_OCTETSTREAM, "application/octet-stream" },
        { 0,    0 }
};

/**
 */
static const tbl_str_itm_t reason_phrase[] = {
        { 200, "OK" },
        { 400, "Bad Request"},
        { 401, "Unauthorized" },
        { 403, "Forbidden" },
        { 404, "Not Found" },
        { 411, "Length Required"},
        { 414, "Request-URI Too Long"},
        { 501, "Not Implemented" },
        { 0,   "" }
};

/*
** This table rows MUST be ordered by size and by alphabetical order
** so we can list througth it quicly
*/
static httpd_content_table_row extension_to_content_tbl[] = {
    {sizeof("js")-1 ,   "js",    CONTENT_TYPE_JS},
    {sizeof("css")-1,   "css",   CONTENT_TYPE_CSS},
    {sizeof("gif")-1,   "gif",   CONTENT_TYPE_GIF}, 
    {sizeof("htm")-1,   "htm",   CONTENT_TYPE_HTML},
    {sizeof("jpg")-1,   "jpg",   CONTENT_TYPE_JPG}, 
    {sizeof("png")-1,   "png",   CONTENT_TYPE_PNG},
    {sizeof("txt")-1,   "txt",   CONTENT_TYPE_PLAIN},
    {sizeof("html")-1,  "html",  CONTENT_TYPE_HTML},
    {sizeof("shtm")-1,  "shtm",  CONTENT_TYPE_HTML},
    {sizeof("shtml")-1, "shtml", CONTENT_TYPE_HTML},
    /* This row MUST have length set to zero so we have proper array termination */
    {0,                      "", CONTENT_TYPE_OCTETSTREAM}
};


/** Send data to socket.
 * \param s < socket
 * \param buf < pointer to data buffer
 * \param len < data length
 * \return length of sent data
 */
static _mqx_int httpd_send(int s, char *buf, int len) {
    int res;

    HTTPD_DEBUG(4, "httpd_send buf:%p len:%d", buf, len);
    res = send(s, buf, len, 0);
    HTTPD_DEBUG(4, "result:%d\n", res);

    return(res);
}

_mqx_int httpd_sendstr(_mqx_int s, char *str)
{
    return httpd_send(s, str, strlen(str));
}

/**
 */
_mqx_int httpd_call_fn(HTTPD_FN_LINK_STRUCT *table, HTTPD_SESSION_STRUCT *session, char *fn_name)
{
    if (table)
    {
        while (*(table->callback))
        {
            if (0 == strcmp(fn_name, table->fn_name))
            {
                table->callback(session);
                break;
            }
            table++;
        }
    }
    return 0;
}

/** Send extended string to socket.
 * \param s < socket
 * \param str < pointer to string
 * \return length of sent data
 */
static _mqx_int httpd_sendextstr(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session, char *str, uint_32 length)
{
    char *src;
    int len, res;
    uint_32 add = 0;
    char fname[HTTPDCFG_MAX_SCRIPT_LN + 1];

    HTTPD_ASSERT(server && session && str);

    fname[0] = 0;

    src = str;

    if (session->response.script_token)
    {
        // script token found
        len = (int)strcspn(src, " ;%<>\r\n\t");

        if (len > 1 && len < HTTPDCFG_MAX_SCRIPT_LN)
        {
            snprintf(fname, len+1, "%s", src);
            // call fn
            HTTPD_DEBUG(2, "script ln: %s\r\n", fname);

            httpd_call_fn(server->params.fn_lnk_tbl, session, fname);
        }

        if (src[len] == '%' && src[len + 1] == '>')
        {
            session->response.script_token = 0;
            len += 1;
        }
        len++;
    }
    else
    {
        for (len = 0; *src && len < length; src++, len++)
        {
            if (*src == '<' && *(src + 1) == '%')
            {
                session->response.script_token = 1;
                src += 2;
                add = 2;
                break;
            }
        }

        res = httpd_send(session->sock, str, len);

        if (res < 0)
        {
            session->response.script_token = 0;
            res = errno;

            if (res != EAGAIN)
            {
                len = 0;
            }  
        }
        else if (len != res)
        {
            session->response.script_token = 0;
        }
    }
    return len+add;
}

_mqx_int httpd_read(HTTPD_SESSION_STRUCT *session, char *dst, _mqx_int len)
{
    int cnt = 0;
    uint_32 buffer_size = strlen(session->buffer.data);
    uint_32 length = (len > buffer_size) ? buffer_size : len;

    /* If there are any data in buffer copy them to user buffer */
    if (buffer_size > 0)
    {
        _mem_copy(session->buffer.data, dst, length);
        memmove(session->buffer.data, session->buffer.data+length, strlen(session->buffer.data+length));
        _mem_zero(session->buffer.data, length);
        session->buffer.offset = 0;
        cnt += length;
    }

    /* If there is some space remaining in user buffer try to read from socket */
    if (cnt < len)
    {
        uint_32 received = 0;
        received = recv(session->sock, dst+cnt, len-cnt, 0);
        
        if (RTCS_ERROR != received)
        {
            cnt += received;
        }
    }
    
    return cnt;
}

/** Get string from table by id
 * \param tbl < pointer to string table
 * \param id < string id
 */
static char* get_tbl_string(tbl_str_itm_t *tbl, const _mqx_int id)
{
    tbl_str_itm_t *ptr = tbl;

    while ((ptr->str) && (id != ptr->id))
    {
        ptr++;
    }
    return ptr->str;
}

void httpd_sendhdr(HTTPD_SESSION_STRUCT *session, _mqx_int content_len, boolean has_entity)
{
    uint_32 length = 0;
    char* buffer = session->buffer.data+session->buffer.offset;
    uint_32 buffer_space = HTTPDCFG_SES_BUFFER_SIZE-session->buffer.offset;

    HTTPD_ASSERT(session);

    if (session->response.hdrsent)
    {
         return;
    }

    length += snprintf(buffer+length, buffer_space-length, "HTTP/1.0 %d %s\r\n", session->response.statuscode,
            get_tbl_string((tbl_str_itm_t*)reason_phrase, session->response.statuscode));

    length += snprintf(buffer+length, buffer_space-length, "%s", "Server: MQX HTTP - Freescale Embedded Web Server\r\n");

    // check authorization
    if (session->response.statuscode == 401)
    {
        length += snprintf(buffer+length, buffer_space-length, "%s", "WWW-Authenticate: Basic realm=\"test\"\n");
    }

    if (session->keep_alive)
    {
        length += snprintf(buffer+length, buffer_space-length, "%s", "Connection: Keep-Alive\r\n");
    }
    else
    {
        length += snprintf(buffer+length, buffer_space-length, "%s", "Connection: close\r\n");
    }
    /* If there will be entity body send content type */
    if (has_entity)
    {
        length += snprintf(buffer+length, buffer_space-length, "Content-Type: %s\r\n", get_tbl_string((tbl_str_itm_t*)content_type, session->response.contenttype));
    }
    
    //only non zero length cause sending Content-Length header field
    if (content_len > 0)
    {
        length += snprintf(buffer+length, buffer_space-length, "Content-Length: %d\r\n", content_len);
    }

    length += snprintf(buffer+length, buffer_space-length, "\r\n");

    httpd_sendstr(session->sock, buffer);
    _mem_zero(buffer, length);
    session->response.hdrsent = 1;

    HTTPD_DEBUG(4, "httpd_sendhdr\n");
}

int httpd_get_contenttype_from_ext(char* extension)
{
    httpd_content_table_row* row = extension_to_content_tbl;
    uint_32 n = 0;
    uint_32 length = 0;

    if (extension != NULL)
    {
        length = strlen(extension);
    }

    /* List throught table rows until length match */
    while ((row->length) && (row->length < length))
    {
        row = extension_to_content_tbl+n;
        n++;
    }

    /* Do a search in valid rows */
    while (row->length == length)
    {
        if (strcasecmp(extension, row->ext) == 0)
        {
            return(row->content_type);
        }
        row = extension_to_content_tbl+n;
        n++;
    }

    /* If we cant find right length or there is no extension return content string from last row */
    while (row->length != 0)
    {
        row = extension_to_content_tbl+n;
        n++;
    }
    return(row->content_type);
}

void httpd_sendfile(HTTPD_STRUCT *server, HTTPD_SESSION_STRUCT *session)
{
    char *ext;
    int expand = 0;
    int len;
    char* buf = session->buffer.data;
    char* temp = NULL;
    
    HTTPD_ASSERT(server && session);

    HTTPD_DEBUG(3, "httpd_sendfile(%p) %s\n", session, session->request.path);

    ext = strrchr(session->request.path, '.');

    if (ext != NULL)
    {
        session->response.contenttype = httpd_get_contenttype_from_ext(ext+1);
    }
    else
    {
        session->response.contenttype = CONTENT_TYPE_OCTETSTREAM;
    }
    
    /* Check if file has server side includes */
    if ((0 == strcasecmp(ext, ".shtml")) || (0 == strcasecmp(ext, ".shtm")))
    {
        expand = 1;
        /* Disable keep-alive for this session otherwise we would have to wait for session timeout */
        session->keep_alive = 0;
        /* If there will be any expansion there's no way how to calculate correct length
        ** zero length prevents sending Content-Length header field. */
        httpd_sendhdr(session, 0, 1);
    }
    else
    {
        httpd_sendhdr(session, session->response.file->SIZE, 1);
    }

    if (expand)
    {
        len = read(session->response.file, buf, HTTPDCFG_SES_BUFFER_SIZE-1);

        if (len > 0)
        {
            len = httpd_sendextstr(server, session, buf, len);
            if (!len)
            {
                session->state = HTTPD_SES_END_REQ;
            }
            else
            {
                session->response.len += len;
                fseek(session->response.file, session->response.len, IO_SEEK_SET);
            }
        }
        else
        {
            session->state = HTTPD_SES_END_REQ;
        }    
    }
    else
    {
        fseek(session->response.file, session->response.len, IO_SEEK_SET);
        len = read(session->response.file, buf, HTTPDCFG_SES_BUFFER_SIZE);
        
        HTTPD_DEBUG(4, "httpd_sendfile READ END %d\n", len);
        
        if (len > 0)
        {
            len = httpd_send(session->sock, buf, len);

            if (len < 0)
            {
                if (errno != EAGAIN)
                {
                    session->state = HTTPD_SES_END_REQ;
                }       
            }
            else
            {
                session->response.len += len;
            }
        }
        else
        {
             session->state = HTTPD_SES_END_REQ;
        }    
    }
    HTTPD_DEBUG(4, "httpd_sendfile END %s \n", session->request.path);
}

void httpd_send_err_page(HTTPD_SESSION_STRUCT *session, const char* title, const char* text)
{
    uint_32 length;
    char* page;
    
    length = snprintf(NULL, 0, ERR_PAGE_FORMAT, title, text);
    length++;
    page = _mem_alloc(length*sizeof(char));

    session->response.contenttype = CONTENT_TYPE_HTML;

    if (page != NULL)
    {
        snprintf(page, length, ERR_PAGE_FORMAT, title, text);
        httpd_sendhdr(session, strlen(page), 1);
        httpd_sendstr(session->sock, page);
        _mem_free(page);
    }
    else
    {
        httpd_sendhdr(session, 0, 0);
    }
}