/**HEADER********************************************************************
* 
* Copyright (c) 2008 Freescale Semiconductor;
* All Rights Reserved                       
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: httpd_task.c$
* $Version : 3.8.12.0$
* $Date    : Aug-31-2012$
*
* Comments:
*
*   HTTPD os specific part implementation.
*
*END************************************************************************/

#include "httpd.h"
#include "httpd_prv.h"
#include "socket.h"

static LWSEM_STRUCT sem_session_counter;

static void httpd_session_task(pointer init_ptr, pointer creator) 
{
    HTTPD_SES_TASK_PARAM *param = (HTTPD_SES_TASK_PARAM*)init_ptr;
    HTTPD_SESSION_STRUCT *session;
    HTTPD_STRUCT* server = param->server;
    uint_32 i = 0;
    boolean found = FALSE;
    _task_id tid =  _task_get_id();
    _mqx_uint retval = 0;

    /* Transfer ownership of parameter from server to session so we can free it */
    retval = _mem_transfer(init_ptr, server->server_tid, tid); 
    if (retval != MQX_OK)
    {
        HTTPD_DEBUG(5, "Failed to transfer task parameter ownership. Error 0x%X.\n", retval);
        RTCS_task_resume_creator(creator, retval);
        _task_destroy(MQX_NULL_TASK_ID);
    }

    HTTPD_ASSERT(init_ptr);
    HTTPD_DEBUG(1, "session task start\n");

    /* Find empty session */
    _lwsem_wait(&server->tid_sem);

    for (i = 0; i < server->params.max_ses; i++)
    {
        if (server->session[i] == NULL)
        {
            found = TRUE;
            break;
        }
    }
     
    if (!found)
    {
        RTCS_task_resume_creator(creator, (uint_32) RTCS_ERROR);
        _mem_free(init_ptr);
        init_ptr = NULL;
        _lwsem_post(&server->tid_sem);
        _lwsem_post(&sem_session_counter);
        _task_destroy(MQX_NULL_TASK_ID);
    }
    
    /* Save task ID - used for indication of running task */
    server->ses_tid[i] = tid;
    /* Access to array done. Unblock other tasks. */
    _lwsem_post(&server->tid_sem);
    
    /* Allocate session */
    session = httpd_ses_alloc(server);
        
    if (session) 
    {
        server->session[i] = session;
        RTCS_task_resume_creator(creator, RTCS_OK);
        
        HTTPD_DEBUG(1, "session task run...\n");
        
        httpd_ses_init(server, session, param->sock);
    
        /* Disable keep-alive for last session so we have at least one session free (not blocked by keep-alive timeout) */
        if (i == server->params.max_ses - 1)
        {
            session->keep_alive_enabled = FALSE;
        }

        while (HTTPD_SESSION_VALID == session->valid) 
        {
            /* Run state machine for session */
            httpd_ses_process(server, session);
            _sched_yield();
        }
        /* If session is not closed close it */
        if (session->state != HTTPD_SES_CLOSE)
        {
            httpd_ses_close(session);
        }
        httpd_ses_free(session);
        server->session[i] = NULL;
    }
    else 
    {
        RTCS_task_resume_creator(creator, (uint_32) RTCS_ERROR);
    }
    /* Cleanup and end task */
    _lwsem_post(&sem_session_counter);
    _mem_free(init_ptr);
    init_ptr = NULL;
    /* Null tid => task is no longer running */
    server->ses_tid[i] = 0;
    HTTPD_DEBUG(1, "session task stop\n");
}


/** server task - httpd main task which create new task for each new request
 */
void httpd_server_task(pointer init_ptr, pointer creator) 
{
    HTTPD_STRUCT   *server = (HTTPD_STRUCT*)init_ptr;
    int            s;
    unsigned short len;
    uint_32        connsock; //socket with connection
    struct         sockaddr sin;

    HTTPD_SES_TASK_PARAM *param = NULL;
    _mqx_uint res;
    uint_32 stack;
    
    HTTPD_DEBUG(1, "server task start\n");
    server->server_tid = _task_get_id();

    res = _lwsem_create(&sem_session_counter, server->params.max_ses);

    if ((!server) && (res != MQX_OK))
    {
        goto ERROR;
    }
          
    RTCS_task_resume_creator(creator, RTCS_OK);
    
    HTTPD_DEBUG(1, "server task run...\n");

    while (server->run) 
    {
        // limit maximal opened sessions
        _lwsem_wait(&sem_session_counter);
        // allocate session task parameter structure
        param = _mem_alloc(sizeof(HTTPD_SES_TASK_PARAM));
        if (param) 
        {
            param->server = server;
            /* Get socket with incoming connection IPv4 or IPv6 */
            connsock = RTCS_selectset(&(server->sock_v4), 2, 0);
            param->sock = accept(connsock, (sockaddr *) &sin, &len);

            if (param->sock != RTCS_SOCKET_ERROR)
            {
                // try create task for session
                res = RTCS_task_create("httpd session", server->params.ses_prio, server->params.ses_stack, httpd_session_task, param);

                if (MQX_OK != res)
                {
                    // session task creation failed, clean and wait...
                    shutdown(param->sock, FLAG_ABORT_CONNECTION);   // abort opened connection
                    _mem_free(param);
                    _lwsem_post(&sem_session_counter);
                    _sched_yield();
                }
            }
            else
            {
                // accept return error, clean and wait, then try again...
                _mem_free(param);
                _lwsem_post(&sem_session_counter);
                /* We probably run out of sockets. Wait some time than try again to prevent session tasks resource starvation */
                _time_delay(150);
            }
        }
        else
        {
            // allocation failed, wait some time and try again
            _time_delay(100);
        }
    }
    
    HTTPD_DEBUG(1, "server task stop\n");
    
ERROR:
    _lwsem_destroy(&sem_session_counter);
    param->server->server_tid = 0;
    RTCS_task_resume_creator(creator, (uint_32)RTCS_ERROR);
}
